__all__ = ['kway_ncut', 'axis_align']

import torch
import torch.nn.functional as F

from ncut_pytorch.utils.sample import farthest_point_sampling
from ncut_pytorch.utils.device import auto_device
from ncut_pytorch.utils.math import chunked_matmul


def kway_ncut(eigvec: torch.Tensor, device: str = None, **kwargs):
    """
    Args:
        eigvec (torch.Tensor): eigenvectors from Ncut output, shape (n, k)
    Returns:
        torch.Tensor: eigenvectors, shape (n, k).
            eigvec.argmax(dim=1) is the cluster assignment.
            eigvec.argmax(dim=0) is the cluster centroids.
    """
    R = axis_align(eigvec, device=device, **kwargs)
    device = auto_device(eigvec.device, device)
    # below is equivalent to: eigvec = eigvec @ R
    eigvec = chunked_matmul(eigvec, R, device=device, large_device=eigvec.device)
    return eigvec


@torch.no_grad()
def axis_align(eigvec: torch.Tensor, device: str = None, max_iter=1000, n_sample=10240, sample_idx=None):
    """Multiclass Spectral Clustering, SX Yu, J Shi, 2003

    Args:
        eigvec (torch.Tensor): continuous eigenvectors from NCUT, shape (n, k)
        max_iter (int, optional): Maximum number of iterations.
        n_sample (int, optional): Number of data points to sample.
    Returns:
        torch.Tensor: Rotation matrix, shape (k, k).
    """

    # subsample the eigenvectors, to speed up the computation
    n, k = eigvec.shape
    if sample_idx is None:
        sample_idx = farthest_point_sampling(eigvec, n_sample, device=device)
    eigvec = eigvec[sample_idx]

    eigvec = F.normalize(eigvec, dim=1)

    # Initialize R matrix with the first column from Farthest Point Sampling
    _sample_idx = farthest_point_sampling(eigvec, k, device=device)
    R = eigvec[_sample_idx].T
    
    original_device = eigvec.device
    original_dtype = eigvec.dtype
    device = auto_device(original_device, device)
    eigvec = eigvec.to(device=device, dtype=torch.float32)
    R = R.to(device=device, dtype=torch.float32)
    
    # Iterative optimization loop
    last_objective_value = 0
    exit_loop = False
    nb_iterations_discretisation = 0

    while not exit_loop:
        nb_iterations_discretisation += 1

        # Discretize the projected eigenvectors
        _eigenvectors_continuous = eigvec @ R
        _eigenvectors_discrete = _onehot_discretize(_eigenvectors_continuous)
        _eigenvectors_discrete = _eigenvectors_discrete.to(device=device, dtype=eigvec.dtype)

        # SVD decomposition
        _out = _eigenvectors_discrete.T @ eigvec
        # Handle autocast for SVD - SVD operations don't support half precision
        _out_dtype = _out.dtype
        try:
            with torch.autocast(device_type=_out.device.type, enabled=False):
                if _out_dtype == torch.float16 or _out_dtype == torch.bfloat16:
                    _out = _out.float()
                U, S, Vh = torch.linalg.svd(_out, full_matrices=False)
        except RuntimeError:
            if _out_dtype == torch.float16 or _out_dtype == torch.bfloat16:
                _out = _out.float()
            U, S, Vh = torch.linalg.svd(_out, full_matrices=False)
        # Convert back to original dtype
        U = U.to(_out_dtype)
        S = S.to(_out_dtype)
        Vh = Vh.to(_out_dtype)
        V = Vh.T

        # Compute the Ncut value
        ncut_value = 2 * (n - torch.sum(S))

        # Check for convergence
        if torch.abs(ncut_value - last_objective_value) < torch.finfo(
            torch.float32).eps or nb_iterations_discretisation > max_iter:
            exit_loop = True
        else:
            last_objective_value = ncut_value
            R = V @ U.T
            
    R = R.to(device=original_device, dtype=original_dtype)
    
    # sort the columns of R by the 2nd row of R
    R = R[:, torch.argsort(R[1])]
    return R


def _onehot_discretize(eigvec):
    _, max_idx = torch.max(eigvec, dim=1)
    eigvec = torch.nn.functional.one_hot(max_idx, num_classes=eigvec.shape[1])
    return eigvec  # (n, k) one-hot vector


from .api import hires_dino_256
from .api import hires_dino_512
from .api import hires_dino_1024
from .api import hires_dinov2

from .hires_dino import HighResDINO
from .lowres_dino import LowResDINO



from ncut_pytorch.color import (
    mspace_color,
    tsne_color,
    umap_color,
    umap_sphere_color,
    convert_to_lab_color,
    rotate_rgb_cube,
)
from ncut_pytorch.ncuts import (
    kway_ncut,
    axis_align,
    ncut_click_prompt,
    ncut_fn
)
from .ncut import Ncut



import torch


def auto_device(existing_device="", user_input_device="auto"):
    """
    Automatically select the best available device or use user-specified device.

    Args:
        existing_device: The device where the features are currently located
        user_input_device: User-specified device or 'auto' for automatic selection

    Returns:
        str: Selected device as a string
    """
    # If user specified a device and it's not 'auto', try to use it
    if user_input_device is not None and str(user_input_device) != "auto":
        try:
            torch.device(str(user_input_device))
            return str(user_input_device)
        except RuntimeError:
            raise ValueError(f"Invalid device: {user_input_device}")

    def prioritize_existing_device(existing_device, new_device):
        existing_device = str(existing_device)
        if new_device in existing_device:
            return existing_device
        return new_device

    # Check for CUDA (NVIDIA GPUs)
    is_cuda_available = torch.cuda.is_available()
    if is_cuda_available:
        return prioritize_existing_device(existing_device, "cuda")

    # Check for MPS (Apple Silicon GPUs)
    try:
        if hasattr(torch, 'mps') and torch.backends.mps.is_available():
            return prioritize_existing_device(existing_device, "mps")
    except (ImportError, AttributeError):
        pass

    # Check for TPU (Google TPUs)
    try:
        import torch_xla.core.xla_model as xm
        return prioritize_existing_device(existing_device, xm.xla_device())
    except (ImportError, AttributeError):
        pass

    # Check for AMD GPUs
    try:
        if hasattr(torch, 'hip') and torch.hip.is_available():
            return prioritize_existing_device(existing_device, "hip")
    except (ImportError, AttributeError):
        pass

    # Check for Intel GPUs
    try:
        if hasattr(torch, 'xpu') and torch.xpu.is_available():
            return prioritize_existing_device(existing_device, "xpu")
    except (ImportError, AttributeError):
        pass

    # Fallback to CPU
    return "cpu"



import torch

from .math import rbf_affinity
from .sample import farthest_point_sampling

@torch.no_grad()
def find_gamma_by_degree(X, d_gamma='auto', affinity_fn=rbf_affinity, X2=None, init_gamma=0.5, r_tol=1e-2, max_iter=100):
    # X: (n_samples, n_features)
    # d_gamma: target mean edge weight
    # binary search for optimal gamma, such that the mean edge weight is close to d_gamma
    if isinstance(d_gamma, float):
        assert d_gamma > 0, "d_gamma must be positive"
    gamma = init_gamma
    scale_inv_gamma = gamma * X.var(0).sum()
    
    # find d_gamma
    if d_gamma == 'auto' or d_gamma is None:
        current_degrees = affinity_fn(X, X2=X2, gamma=scale_inv_gamma).mean(1)
        current_degree = current_degrees.mean().item()
        mask = current_degrees < current_degree
        d_gamma = bin_and_find_mode(current_degrees[mask])
    
    # find gamma, binary search
    current_degree = affinity_fn(X, X2=X2, gamma=gamma).mean().item()
    i_iter = 0
    low, high = 0, float('inf')
    tol = r_tol * d_gamma
    while abs(current_degree - d_gamma) > tol and i_iter < max_iter:
        if current_degree > d_gamma:
            high = gamma
            gamma = (low + gamma) / 2
        else:
            low = gamma
            gamma = gamma * 2 if high == float('inf') else (gamma + high) / 2
        current_degree = affinity_fn(X, X2=X2, gamma=gamma).mean().item()
        i_iter += 1
        
    return gamma

@torch.no_grad()
def find_gamma_by_degree_after_fps(X, d_gamma='auto', affinity_fn=rbf_affinity, n_sample=1000, **kwargs):
    indices = farthest_point_sampling(X, n_sample)
    return find_gamma_by_degree(X[indices], d_gamma, affinity_fn, **kwargs)


@torch.no_grad()
def bin_and_find_mode(degrees, n_bins=20):
    degrees = degrees.to(torch.float32).to(torch.device('cpu'))
    counts, bin_edges = torch.histogram(degrees, bins=n_bins)
    
    middle_bin_idx = n_bins // 2
    
    # Get the actual data in that bin
    left_edge = bin_edges[middle_bin_idx]
    right_edge = bin_edges[middle_bin_idx + 1]
    
    mask = (degrees >= left_edge) & (degrees < right_edge)
    
    # Return the mean of the most populous bin
    return degrees[mask].mean().item()


from .coloring import mspace_color
from .coloring import tsne_color
from .coloring import umap_color
from .coloring import umap_sphere_color
from .coloring import convert_to_lab_color
from .coloring import rotate_rgb_cube


# Credits: https://github.com/tldr-group/HR-Dv2

from functools import partial
from types import MethodType
from typing import List, Tuple, Literal, Union

# ==================== IMPORTS ====================
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.modules.utils import _pair

from .patch import Patch
from .transform import combine_transforms_pairwise
from .transform import iden_partial, get_shift_transforms, get_flip_transforms

Interpolation = Literal[
    "nearest", "linear", "bilinear", "bicubic", "trilinear", "area", "nearest-exact"
]
AttentionOptions = Literal["q", "k", "v", "o", "none"]
DINONameOptions = Literal["dino_vitb8", "dino_vits8", "dino_vitb16", "dinov2_vitb14", "dinov2_vitb14_reg"]

# ==================== MODULE ====================


class HighResDINO(nn.Module):
    def __init__(
        self,
        dino_name: DINONameOptions,
        stride: int = 5,
        dtype: Union[torch.dtype, int] = torch.float16,
        track_grad: bool = False,
        attention_mask_ratio: float = 0.1,
        chunk_size: int = 4,
        feature_resolution: int = 1024,
    ) -> None:
        super().__init__()

        self.dinov2: nn.Module
        if "dinov2" in dino_name:
            hub_path = "facebookresearch/dinov2"
            self.dinov2 = torch.hub.load(hub_path, dino_name, verbose=False)
        elif "dino" in dino_name:
            hub_path = "facebookresearch/dino:main"
            self.dinov2 = torch.hub.load(hub_path, dino_name, verbose=False)
        else:
            raise ValueError(f"Invalid model name: {dino_name}")

        self.dinov2.eval()

        if "dinov2" not in dino_name:
            self.dinov2.num_heads = 6  # type: ignore
            self.dinov2.num_register_tokens = 0  # type: ignore

        # Get params of Dv2 model and store references to original settings & methods
        feat, patch = self.get_model_params(dino_name)
        self.original_patch_size: int = patch
        self.original_stride = _pair(patch)
        # may need to deepcopy this instead of just referencing
        # self.original_pos_enc = self.dinov2.interpolate_pos_encoding
        self.feat_dim: int = feat
        self.n_heads: int = 6
        self.n_register_tokens = 4
        self.chunk_size = chunk_size
        self.feature_resolution = feature_resolution
        
        # mask out some of the attention Keys (K) to save compute
        self.attention_mask_ratio = attention_mask_ratio

        self.stride = _pair(stride)
        # we need to set the stride to the original once before we set it to desired stride
        # i don't know why
        self.set_model_stride(self.dinov2, patch)
        self.set_model_stride(self.dinov2, stride)

        self.transforms: List[partial] = []
        self.inverse_transforms: List[partial] = []
        self.interpolation_mode: Interpolation = "nearest-exact"

        # If we want to save memory, change to float16
        if type(dtype) == int:
            dtype = torch.float16 if dtype == 16 else torch.float32

        self.dtype = dtype
        if dtype != torch.float32:
            self = self.to(dtype)
        self.track_grad = track_grad  # off by default to save memory

        self.patch_last_block(self.dinov2, dino_name)

    def get_model_params(self, dino_name: str) -> Tuple[int, int]:
        """Match a name like dinov2_vits14 / dinov2_vitg16_lc etc. to feature dim and patch size.

        :param dino_name: string of dino model name on torch hub
        :type dino_name: str
        :return: tuple of original patch size and hidden feature dimension
        :rtype: Tuple[int, int]
        """
        split_name = dino_name.split("_")
        model = split_name[1]
        arch, patch_size = model[3], int(model[4:])
        feat_dim_lookup = {"s": 384, "b": 768, "l": 1024, "g": 1536}
        feat_dim: int = feat_dim_lookup[arch]
        return feat_dim, patch_size

    def set_model_stride(
        self, dino_model: nn.Module, stride_l: int, verbose: bool = False
    ) -> None:
        """Create new positional encoding interpolation method for $dino_model with
        supplied $stride, and set the stride of the patch embedding projection conv2D
        to $stride.

        :param dino_model: Dv2 model
        :type dino_model: DinoVisionTransformer
        :param new_stride: desired stride, usually stride < original_stride for higher res
        :type new_stride: int
        :return: None
        :rtype: None
        """

        new_stride_pair = torch.nn.modules.utils._pair(stride_l)
        # if new_stride_pair == self.stride:
        #    return  # early return as nothing to be done
        self.stride = new_stride_pair
        dino_model.patch_embed.proj.stride = new_stride_pair  # type: ignore
        if verbose:
            print(f"Setting stride to ({stride_l},{stride_l})")

        # if new_stride_pair == self.original_stride:
        # if resetting to original, return original method
        #    dino_model.interpolate_pos_encoding = self.original_pos_enc  # type: ignore
        # else:
        dino_model.interpolate_pos_encoding = MethodType(  # type: ignore
            Patch._fix_pos_enc(self.original_patch_size, new_stride_pair),
            dino_model,
        )  # typed ignored as they can't type check reassigned methods (generally is poor practice)

    def patch_last_block(self, dino_model: nn.Module, dino_name: str) -> None:
        """Patch the final block of the dino model to add attention return code.

        :param dino_model: DINO or DINOv2 model
        :type dino_model: nn.Module
        """
        final_block = dino_model.blocks[-1]  # type: ignore
        attn_block = final_block.attn  # type: ignore
        # hilariously this also works for dino i.e we can patch dino's attn block forward to
        # use the memeory efficienty attn like in dinov2
        attn_block.forward = MethodType(Patch._fix_attn_masking(self.attention_mask_ratio), attn_block)
        if "dinov2" in dino_name:
            final_block.forward = MethodType(Patch._fix_block_forward_dv2(), final_block)  # type: ignore
            dino_model.forward_feats_attn = MethodType(  # type: ignore
                Patch._add_new_forward_features_dv2(), dino_model
            )
        elif "dino" in dino_name:
            for i, blk in enumerate(dino_model.blocks):
                blk.forward = MethodType(Patch._fix_block_forward_dino(), blk)
                attn_block = blk.attn
                attn_block.forward = MethodType(Patch._fix_attn_masking(self.attention_mask_ratio), attn_block)
            final_block.forward = MethodType(Patch._fix_block_forward_dino(), final_block)  # type: ignore
            dino_model.forward_feats_attn = MethodType(  # type: ignore
                Patch._add_new_forward_features_dino(), dino_model
            )
        else:
            for i, blk in enumerate(dino_model.blocks):
                blk.forward = MethodType(Patch._fix_block_forward_dino(), blk)
                attn_block = blk.attn
                attn_block.forward = MethodType(Patch._fix_attn_masking(self.attention_mask_ratio), attn_block)
            final_block.forward = MethodType(Patch._fix_block_forward_dino(), final_block)  # type: ignore
            dino_model.forward_feats_attn = MethodType(  # type: ignore
                Patch._add_new_forward_features_vit(), dino_model
            )

    def get_n_patches(self, img_h: int, img_w: int) -> Tuple[int, int]:
        stride_l = self.stride[0]
        n_patch_h: int = 1 + (img_h - self.original_patch_size) // stride_l
        n_patch_w: int = 1 + (img_w - self.original_patch_size) // stride_l
        return (n_patch_h, n_patch_w)

    def set_transforms(
        self, transforms: List[partial], inv_transforms: List[partial]
    ) -> None:
        assert len(transforms) == len(
            inv_transforms
        ), "Each transform must have an inverse!"
        self.transforms = transforms
        self.inverse_transforms = inv_transforms

    @torch.no_grad()
    def get_transformed_input_batch(
        self, x: torch.Tensor, transforms: List[partial]
    ) -> torch.Tensor:
        """Loop through a list of (invertible) transforms, apply them to input $x, store
        in a list then batch and return.

        :param x: input unbatched standardized image tensor
        :type x: torch.Tensor
        :param transforms: list of partial functions representing transformations on image
        :type transforms: List[partial]
        :return: batch of transformed images
        :rtype: torch.Tensor
        """
        img_list: List[torch.Tensor] = []
        if len(transforms) == 0:  # if we want to test VE by itself
            # print("Warning: no transforms supplied, using identity transform")
            self.transforms.append(iden_partial)
            self.inverse_transforms.append(iden_partial)

        for transform in transforms:
            transformed_img: torch.Tensor = transform(x)
            img_list.append(transformed_img)

        if len(img_list[0].shape) == 3:  # if c, h, w then stack list
            img_batch = torch.stack(img_list)
        else:  # if b, c, h, w then cat
            img_batch = torch.cat(img_list)

        if self.dtype != torch.float32:
            img_batch = img_batch.to(self.dtype)
        return img_batch

    @torch.no_grad()
    def invert_transforms(
        self, feature_batch: torch.Tensor, x: torch.Tensor
    ) -> torch.Tensor:
        """For each flat Dv2 features of our transformed imgs in $feature_batch, loop through,
        make them spatial again by reshaping, permuting and resizing, then perform the
        corresponding inverse transform and add to our summand variable. Finally we divide by
        N_imgs to create average.

        # TODO: parameterise this with the inverse transform s.t can just feed single batch (of
        say the attn map) and the iden partial transform and get upsampled

        :param feature_batch: batch of N_transform features from Dv2 with shape (n_patches, n_features)
        :type feature_batch: torch.Tensor
        :param x: original input img of size (channels, img_h, img_w), useful for resizing later
        :type x: torch.Tensor
        :return: feature image of size (n_features, img_h, img_w)
        :rtype: torch.Tensor
        """
        _, img_h, img_w = x.shape
        c = feature_batch.shape[-1]
        stride_l = self.stride[0]
        n_patch_w: int = 1 + (img_w - self.original_patch_size) // stride_l
        n_patch_h: int = 1 + (img_h - self.original_patch_size) // stride_l

        # Summand variable here to be memory efficient
        out_feature_img: torch.Tensor = torch.zeros(
            1,
            c,
            self.feature_resolution,
            self.feature_resolution,
            device=x.device,
            dtype=self.dtype,
            requires_grad=self.track_grad,
        )

        for i, inv_transform in enumerate(self.inverse_transforms):
            feat_patch_flat = feature_batch[i]
            # interp expects batched spatial tensors so reshape and unsqueeze
            feat_patch = feat_patch_flat.view((n_patch_h, n_patch_w, c))

            permuted = feat_patch.permute((2, 0, 1)).unsqueeze(0)

            full_size = F.interpolate(
                permuted,
                (img_h, img_w),
                mode=self.interpolation_mode,
            )
            inverted: torch.Tensor = inv_transform(full_size)
            
            # resize the inverted feature map to the output resolution
            out = F.interpolate(
                inverted, 
                (self.feature_resolution, self.feature_resolution), 
                mode=self.interpolation_mode
            )
            
            out_feature_img += out

        n_imgs: int = feature_batch.shape[0]
        mean = out_feature_img / n_imgs
        mean = mean.squeeze(0)
        return mean

    def _forward_one_image(
        self,
        x: torch.Tensor,
        attn_choice: AttentionOptions = "none",
    ) -> torch.Tensor:
        """Feed input img $x through network and get high-res features.

        :param x: unbatched image tensor, (c, h, w)
        :param attn_choice: choice of attention, "none" or "q", "k", "v", "o"
        :return: upsampled features, (c, h, w)
        """
        if self.dtype != torch.float32:  # cast (i.e to f16)
            x = x.type(self.dtype)

        img_batch = self.get_transformed_input_batch(x, self.transforms)
        N_imgs = img_batch.shape[0]
        
        all_features = []
        for i in range(0, N_imgs, self.chunk_size):
            _img_batch = img_batch[i:i+self.chunk_size]
            out_dict = self.dinov2.forward_feats_attn(_img_batch, None, attn_choice)  # type: ignore
            if attn_choice != "none":
                feats, attn = out_dict["x_norm_patchtokens"], out_dict["x_patchattn"]
                features_batch = torch.concat((feats, attn), dim=-1)
            else:
                features_batch = out_dict["x_norm_patchtokens"]

            if self.dtype != torch.float32:  # cast (i.e to f16)
                features_batch = features_batch.type(self.dtype)
            
            all_features.append(features_batch)

        features_batch = torch.cat(all_features, dim=0)
        upsampled_features = self.invert_transforms(features_batch, x)
        return upsampled_features
    
    def _forward_auto_chunk(self, *args, **kwargs) -> torch.Tensor:
        try:
            return self._forward_one_image(*args, **kwargs)
        except torch.cuda.OutOfMemoryError as e:
            raise RuntimeError("CUDA out of memory, please try to reduce the chunk size.\n    hires_dino_model.chunk_size = 1")
        
    def forward(
        self,
        x: torch.Tensor,
        attn_choice: AttentionOptions = "none",
        move_to_cpu: bool = True,
    ) -> torch.Tensor:
        """Feed input img $x through network and get low and high res features.

        :param x: batched image tensor, (b, c, h, w)
        :param attn_choice: choice of attention, "none" or "q", "k", "v", "o"
        :return: upsampled features, (b, c, h, w)
        :rtype: torch.Tensor
        """
        upsampled_features = []
        for i in range(x.shape[0]):
            if self.track_grad:
                out = self._forward_auto_chunk(x[i], attn_choice)
            else:
                with torch.no_grad():
                    out = self._forward_auto_chunk(x[i], attn_choice)
                    if move_to_cpu:
                        out = out.cpu()
            upsampled_features.append(out)
        upsampled_features = torch.stack(upsampled_features, dim=0)
        return upsampled_features

    

def hires_dino(dino_name: DINONameOptions = "dino_vitb8",
               stride: int = 6,
               shift_dists: List[int] = [1, 2, 3],
               flip_transforms: bool = True,
               attention_mask_ratio: float = 0.1,
               dtype: Union[torch.dtype, int] = torch.float16,
               track_grad: bool = False,
               chunk_size: int = 4,
               feature_resolution: int = 512
               ) -> HighResDINO:
    """
    Args:
        dino_name: name of the DINO model to use
        stride: stride size of the tokenization, smaller is better but slower
        shift_dists: pixel shifts for multiple image transformations, more shifts means more crispy features
        flip_transforms: whether to use flip transforms, remove positional features
        attention_mask_ratio: ratio of attention keys to mask out
        dtype: data type of the model
        track_grad: whether to track gradients
        chunk_size: number of images to process in one batch, in case of OOM
        feature_resolution: resolution of the output features
    """

    model = HighResDINO(dino_name, stride, dtype, track_grad, attention_mask_ratio, chunk_size, feature_resolution)

    fwd_shift, inv_shift = get_shift_transforms(shift_dists)
    if flip_transforms:  # add flip transforms
        fwd_flip, inv_flip = get_flip_transforms()
        fwd, inv = combine_transforms_pairwise(fwd_shift, fwd_flip, inv_shift, inv_flip)
    else:
        fwd, inv = fwd_shift, inv_shift
    model.set_transforms(fwd, inv)

    model = model.eval()

    return model





import math

import numpy as np
import torch

from .torch_mod import svd_lowrank as my_torch_svd_lowrank


def rbf_affinity(
        X1: torch.Tensor,
        X2: torch.Tensor = None,
        gamma: float = 1.0,
):
    """Compute affinity matrix from input features.

    Args:
        X1 (torch.Tensor): input features, shape (n_samples, n_features)
        X2 (torch.Tensor, optional): optional, if not None, compute affinity between two features
        gamma (float): affinity matrix parameter, lower t reduce the edge weights
            on weak connections, default 1.0

    Returns:
        (torch.Tensor): affinity matrix, shape (n_samples, n_samples)
    """
    X2 = X1 if X2 is None else X2

    distances = torch.cdist(X1, X2, p=2) ** 2
    A = torch.exp(-distances / (2 * gamma + 1e-8))
    return A

def cosine_affinity(
    X1: torch.Tensor, 
    X2: torch.Tensor = None, 
    gamma: float = 1.0,
    repluse: bool = False,
):
    """Compute cosine similarity affinity matrix from input features.
    """
    X2 = X1 if X2 is None else X2
    X1_norm = torch.nn.functional.normalize(X1, p=2, dim=1, eps=1e-8)
    X2_norm = torch.nn.functional.normalize(X2, p=2, dim=1, eps=1e-8)
    S = torch.mm(X1_norm, X2_norm.T)
    if not repluse:
        num = S - 1
    if repluse:
        num = S + 1
    A = torch.exp(- num**2 / (2 * gamma + 1e-8)**2)
    return A

def keep_topk_per_row(A: torch.Tensor, k: int = 10):
    """
    Args:
        A (torch.Tensor): affinity matrix, shape (n_samples, n_samples)
        k (int): number of topk values to return
    Returns:
        (torch.Tensor): topk values, shape (n_samples, k)
    """
    topk_A, topk_indices = A.topk(k=k, dim=-1, largest=True)
    return topk_A, topk_indices

def svd_lowrank(mat: torch.Tensor, q: int):
    """
    SVD lowrank
    mat: (n, m), n data, m features
    q: int
    return: (n, q), (q,), (q, m)
    """
    dtype = mat.dtype

    try:
        with torch.autocast(device_type=mat.device.type, enabled=False):
            if dtype == torch.float16 or dtype == torch.bfloat16:
                mat = mat.float()  # svd_lowrank does not support float16

            u, s, v = my_torch_svd_lowrank(mat, q=q + 10)
    except RuntimeError:
        if dtype == torch.float16 or dtype == torch.bfloat16:
            mat = mat.float()
        u, s, v = my_torch_svd_lowrank(mat, q=q + 10)

    u = u[:, :q]
    s = s[:q]
    v = v[:, :q]

    u = u.to(dtype)
    s = s.to(dtype)
    v = v.to(dtype)
    return u, s, v


def pca_lowrank(mat, q):
    """
    PCA lowrank
    mat: (n, m), n data, m features
    q: int
    return: (n, q), (q,), (q, m)
    """
    u, s, v = svd_lowrank(mat, q)
    _n = mat.shape[0]
    s /= math.sqrt(_n)
    return u @ torch.diag(s)


def quantile_min_max(x, q1=0.01, q2=0.99, n_sample=10000):
    if x.shape[0] > n_sample:
        # random sampling to reduce the load of quantile calculation, torch.quantile does not support large tensor
        np.random.seed(0)
        random_idx = np.random.choice(x.shape[0], n_sample, replace=False)
        vmin, vmax = x[random_idx].quantile(q1), x[random_idx].quantile(q2)
    else:
        vmin, vmax = x.quantile(q1), x.quantile(q2)
    return vmin, vmax


def quantile_normalize(x, q=0.95):
    """normalize each dimension of x to [0, 1], take 95-th percentage, this robust to outliers
        </br> 1. sort x
        </br> 2. take q-th quantile
        </br>     min_value -> (1-q)-th quantile
        </br>     max_value -> q-th quantile
        </br> 3. normalize
        </br> x = (x - min_value) / (max_value - min_value)

    Args:
        x (torch.Tensor): input tensor, shape (n_samples, n_features)
            normalize each feature to 0-1 range
        q (float): quantile, default 0.95

    Returns:
        torch.Tensor: quantile normalized tensor
    """
    # normalize x to 0-1 range, max value is q-th quantile
    # quantile makes the normalization robust to outliers
    if isinstance(x, np.ndarray):
        x = torch.tensor(x)
    vmax, vmin = quantile_min_max(x, q, 1 - q)
    x = (x - vmin) / (vmax - vmin)
    x = x.clamp(0, 1)
    return x


def gram_schmidt(matrix):
    """Orthogonalize a matrix column-wise using the Gram-Schmidt process.

    Args:
        matrix (torch.Tensor): A matrix to be orthogonalized (m x n).
            the second dimension is orthogonalized
    Returns:
        torch.Tensor: Orthogonalized matrix (m x n).
    """

    # Get the number of rows (m) and columns (n) of the input matrix
    m, n = matrix.shape

    # Create an empty matrix to store the orthogonalized columns
    orthogonal_matrix = torch.zeros((m, n), dtype=matrix.dtype, device=matrix.device)

    for i in range(n):
        # Start with the i-th column of the input matrix
        vec = matrix[:, i]

        for j in range(i):
            # Subtract the projection of vec onto the j-th orthogonal column
            proj = torch.dot(orthogonal_matrix[:, j], matrix[:, i]) / torch.dot(
                orthogonal_matrix[:, j], orthogonal_matrix[:, j]
            )
            vec = vec - proj * orthogonal_matrix[:, j]

        # Store the orthogonalized vector
        orthogonal_matrix[:, i] = vec / torch.norm(vec)

    return orthogonal_matrix


def chunked_matmul(
        A,
        B,
        device,
        chunk_size=65536,
        large_device="cpu",
        transform=lambda x: x,
):
    """
    Chunked matrix multiplication, to avoid OOM
    equivalent to: out = A @ B
    """
    A = A.to(large_device)
    B = B.to(large_device)
    C = torch.zeros(A.shape[0], B.shape[1], device=large_device, dtype=A.dtype)
    for i in range(0, A.shape[0], chunk_size):
        end_i = min(i + chunk_size, A.shape[0])
        for j in range(0, B.shape[1], chunk_size):
            end_j = min(j + chunk_size, B.shape[1])
            _A = A[i:end_i]
            _B = B[:, j:end_j]
            _C_ij = None
            for k in range(0, A.shape[1], chunk_size):
                end_k = min(k + chunk_size, A.shape[1])
                __A = _A[:, k:end_k].to(device)
                __B = _B[k:end_k].to(device)
                _C = __A @ __B
                _C_ij = _C if _C_ij is None else _C_ij + _C
            _C_ij = transform(_C_ij)

            _C_ij = _C_ij.to(large_device)
            C[i:end_i, j:end_j] = _C_ij
    return C


def correct_rotation(eigvec):
    # correct the random rotation (flipping sign) of eigenvectors
    with torch.no_grad():
        rand_w = torch.ones(eigvec.shape[0], device=eigvec.device, dtype=eigvec.dtype)
        s = rand_w[None, :] @ eigvec
        s = s.sign()
    eigvec = eigvec * s
    return eigvec


def normalize_affinity(A, eps=1e-8):
    with torch.no_grad():
        D = A.abs().sum(dim=-1)
        D = D + eps
    A = A / torch.sqrt(D)[:, None]
    A = A / torch.sqrt(D)[None, :]
    return A


@torch.no_grad()
def compute_delaunay(points):
    """Compute Delaunay triangulation of points"""
    from scipy.spatial import Delaunay
    if points.shape[1] > 3:
        points = pca_lowrank(points, 3)
    return Delaunay(points.cpu().numpy()).simplices


def compute_riemann_curvature_loss(points, simplices=None, domain_min=0, domain_max=1):
    """
    Calculate loss based on approximated Riemann curvature.
    
    The loss measures deviations from uniform metric tensors across simplices,
    which approximates variations in Riemann curvature.
    """
    if simplices is None:
        simplices = compute_delaunay(points)

    ideal_det = torch.tensor(1.0, device=points.device, dtype=torch.float64)

    # Process each simplex in parallel 
    simplices_tensor = torch.tensor(simplices, device=points.device)

    # Extract points that form each simplex
    simplex_points = points[simplices_tensor]

    # Calculate edge vectors from the first point of each simplex
    edges = simplex_points[:, 1:] - simplex_points[:, 0].unsqueeze(1)

    # Compute metric tensors (Gram matrices) for each simplex
    metric_tensors = torch.matmul(edges, edges.transpose(1, 2))

    # Calculate determinants (related to volume distortion)
    dets = torch.linalg.det(metric_tensors)

    # Penalize deviations from constant determinant
    valid_dets = dets[dets > 0]
    total_curvature = torch.mean((valid_dets - ideal_det) ** 2)
    return total_curvature


def compute_axis_align_loss(points):
    """ Encourage axis alignment by minimizing off-diagonal elements in the covariance matrix """
    n, d = points.shape
    centered_data = points - points.mean(dim=0)  # Center the data
    cov_matrix = (centered_data.T @ centered_data) / n  # Compute covariance matrix

    eye = torch.eye(d, device=points.device)
    return torch.mean((cov_matrix - eye) ** 2)


def compute_repulsion_loss(points):
    dist_matrix = torch.cdist(points, points)
    # Set diagonal to large value to avoid self-repulsion 
    mask = torch.eye(points.shape[0], device=points.device).bool()
    dist_matrix = dist_matrix + mask * 1e10

    # For each point, only consider repulsion from nearest neighbor
    nearest_dists, _ = torch.min(dist_matrix, dim=1)
    repulsion = 1.0 / (nearest_dists + 0.01)  # the shift is to avoid big gradient
    return torch.mean(repulsion)


def compute_attraction_loss(points):
    center = points.mean(dim=0)
    dist = (points - center) ** 2
    return dist.mean()


def compute_boundary_loss(points, domain_min=-1, domain_max=1):
    return torch.mean((torch.relu(domain_min - points)) ** 2) + \
        torch.mean((torch.relu(points - domain_max)) ** 2)


def find_elbow(eigvals, n_elbows=5):
    # Convert to numpy array if tensor
    if torch.is_tensor(eigvals):
        eigvals = eigvals.cpu().detach().numpy()

    # Calculate coordinates
    coords = np.vstack((np.arange(len(eigvals)), eigvals)).T

    # Get vector from first to last point
    line_vec = coords[-1] - coords[0]
    line_vec_norm = line_vec / np.sqrt(np.sum(line_vec ** 2))

    # Vector from point to first point
    vec_from_first = coords - coords[0]

    # Distance from points to line
    dist_from_line = np.cross(line_vec_norm, vec_from_first)

    # Find elbow points (maximum distances)
    elbow_indices = []
    remaining_distances = np.abs(dist_from_line)

    for _ in range(n_elbows):
        if len(remaining_distances) == 0:
            break
        elbow_idx = np.argmax(remaining_distances)
        elbow_indices.append(elbow_idx)
        # Zero out a window around the found elbow to find the next one
        window = 5  # Adjust this window size as needed
        start_idx = max(0, elbow_idx - window)
        end_idx = min(len(remaining_distances), elbow_idx + window)
        remaining_distances[start_idx:end_idx] = 0

    return sorted(elbow_indices)



__all__ = ['ncut_fn', 'nystrom_propagate']

from typing import Callable, Union

import torch

from ncut_pytorch.utils.gamma import find_gamma_by_degree, find_gamma_by_degree_after_fps
from ncut_pytorch.utils.math import rbf_affinity, cosine_affinity
from ncut_pytorch.utils.math import gram_schmidt, normalize_affinity, svd_lowrank, correct_rotation, keep_topk_per_row
from ncut_pytorch.utils.sample import farthest_point_sampling
from ncut_pytorch.utils.device import auto_device
from ncut_pytorch.utils.grad import set_grad_enabled


class NystromConfig:
    """
    Internal configuration for nystrom approximation, can be overridden by kwargs
    Values are optimized based on empirical experiments, no need to change the values
    """
    n_sample = 10240  # number of samples for nystrom approximation, 10240 is large enough for most cases
    n_sample_max_ratio = 1/4  # max ratio of n_sample to n_data, not full sample ensures balanced sampling
    n_sample2 = 1024  # number of samples for eigenvector propagation, 1024 is large enough for most cases
    n_neighbors = 32  # number of neighbors for eigenvector propagation, 10 is large enough for most cases
    n_neighbors_max_ratio = 1/32  # max ratio of n_neighbors to n_sample2, to avoid over smoothing
    matmul_chunk_size = 65536  # chunk size for matrix multiplication, larger chunk size is faster but requires more memory
    move_output_to_cpu = True  # if True, will move output to cpu, saves VRAM
    
    def update(self, kwargs: dict):
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                raise ValueError(f"Invalid kwarg: {key}")


def ncut_fn(
        X: torch.Tensor,
        n_eig: int = 100,
        track_grad: bool = False,
        d_gamma: float = None,
        device: str = None,
        gamma: float = None,
        repulsion_gamma: float = None,
        repulsion_weight: float = 0.2,
        make_orthogonal: bool = False,
        affinity_fn: Union["rbf_affinity", "cosine_affinity"] = rbf_affinity,
        no_propagation: bool = False,
        **kwargs,
) -> Union[tuple[torch.Tensor, torch.Tensor], tuple[torch.Tensor, torch.Tensor, torch.Tensor, float]]:
    """Normalized Cut, balanced sampling and nystrom approximation.

    Args:
        X (torch.Tensor): input features, shape (N, D)
        n_eig (int): number of eigenvectors
        track_grad (bool): keep track of pytorch gradients
        d_gamma (float): affinity gamma parameter, lower d_gamma results in sharper eigenvectors
        device (str): device, default 'auto' (auto detect GPU)
        gamma (float): affinity parameter, override d_gamma if provided
        make_orthogonal (bool): make eigenvectors orthogonal
        affinity_fn (callable): affinity function, default rbf_affinity. Should accept (X1, X2=None, gamma=float) and return affinity matrix
        
    Returns:
        eigenvectors (torch.Tensor): shape (N, n_eig)
        eigenvalues (torch.Tensor): sorted in descending order, shape (n_eig,)
    
    Examples:
        >>> from ncut_pytorch import ncut_fn
        >>> import torch
        >>> features = torch.rand(10000, 100)
        >>> eigvec, eigval = ncut_fn(features, n_eig=20)
        >>> print(eigvec.shape, eigval.shape)  # (10000, 20) (20,)
    """
    config = NystromConfig()
    config.update(kwargs)

    # use GPU if available
    device = auto_device(X.device, device)

    with set_grad_enabled(track_grad):
        # subsample for nystrom approximation
        n_sample = min(config.n_sample, int(X.shape[0]*config.n_sample_max_ratio))
        nystrom_indices = farthest_point_sampling(X, n_sample=n_sample, device=device)
        nystrom_X = X[nystrom_indices].to(device)

        # find optimal gamma for affinity matrix
        if gamma is None:
            if affinity_fn == rbf_affinity:
                gamma = find_gamma_by_degree_after_fps(nystrom_X, d_gamma, affinity_fn)
            else:
                gamma = 0.5
                
        if repulsion_gamma is not None:
            nystrom_eigvec, eigval = ncut_with_repulsion(nystrom_X, n_eig, gamma_attraction=gamma, gamma_repulsion=repulsion_gamma, repulsion_weight=repulsion_weight, affinity_fn=affinity_fn)
        else:
            A = affinity_fn(nystrom_X, gamma=gamma)
            nystrom_eigvec, eigval = _plain_ncut(A, n_eig)

        if no_propagation:
            return nystrom_eigvec, eigval, nystrom_indices, gamma

        # propagate eigenvectors from subgraph to full graph
        eigvec = nystrom_propagate(
            nystrom_eigvec,
            X,
            nystrom_X,
            n_neighbors=config.n_neighbors,
            n_sample=config.n_sample2,
            matmul_chunk_size=config.matmul_chunk_size,
            device=device,
            move_output_to_cpu=config.move_output_to_cpu,
            track_grad=track_grad,
        )

        # post-hoc orthogonalization
        if make_orthogonal:
            eigvec = gram_schmidt(eigvec)

        return eigvec, eigval
    

def ncut_with_repulsion(
    X: torch.Tensor,
    n_eig: int = 100,
    gamma_attraction: float = None,
    gamma_repulsion: float = None,
    repulsion_weight: float = 0.2,
    affinity_fn: Union["rbf_affinity", "cosine_affinity"] = cosine_affinity,
    eps: float = 1e-8,
):
    A = affinity_fn(X, gamma=gamma_attraction)
    R = affinity_fn(X, gamma=gamma_repulsion, repluse=True)
    R = R * repulsion_weight
    D_A = A.sum(1) + eps
    D_R = R.sum(1) + eps
    D = D_A + D_R
    W = A - R + torch.diag(D_R)
    W = W / D[:, None]
    eigvec, eigval, _ = svd_lowrank(W, n_eig)
    eigvec = correct_rotation(eigvec)
    return eigvec, eigval


def _plain_ncut(
        A: torch.Tensor,
        n_eig: int = 100,
):
    # normalization; A = D^(-1/2) A D^(-1/2)
    A = normalize_affinity(A)

    eigvec, eigval, _ = svd_lowrank(A, n_eig)

    eigvec = correct_rotation(eigvec)

    return eigvec, eigval


def nystrom_propagate(
        nystrom_out: torch.Tensor,
        X: torch.Tensor,
        nystrom_X: torch.Tensor,
        track_grad: bool = False,
        device: str = None,
        return_indices: bool = False,
        **kwargs,
) -> Union[torch.Tensor, tuple[torch.Tensor, torch.Tensor]]:
    """propagate output from nystrom sampled nodes to all nodes,
    use a weighted sum of the nearest neighbors to propagate the output.

    Args:
        nystrom_out (torch.Tensor): output from nystrom sampled nodes, shape (m, D)
        X (torch.Tensor): input features for all nodes, shape (N, D)
        nystrom_X (torch.Tensor): input features from nystrom sampled nodes, shape (m, D)
        gamma (float): affinity parameter, default 1.0
        track_grad (bool): keep track of pytorch gradients, default False
        device (str): device to use for computation, if 'auto', will detect GPU automatically
        affinity_fn (callable): affinity function, default rbf_affinity. Should accept (X1, X2=None, gamma=float) and return affinity matrix

    Returns:
        torch.Tensor: output propagated by nearest neighbors, shape (N, D)
    """

    config = NystromConfig()
    config.update(kwargs)

    with set_grad_enabled(track_grad):
        device = auto_device(nystrom_out.device, device)
        indices = farthest_point_sampling(nystrom_out, config.n_sample2, device=device)
        nystrom_out = nystrom_out[indices].to(device)
        nystrom_X = nystrom_X[indices].to(device)
        
        gamma = find_gamma_by_degree(nystrom_X, affinity_fn=rbf_affinity)
        
        D = rbf_affinity(nystrom_X, gamma=gamma).mean(1)

        all_outs = []
        n_chunk = config.matmul_chunk_size
        n_neighbors = int(min(config.n_neighbors, len(indices)*config.n_neighbors_max_ratio))
        n_neighbors = max(n_neighbors, 4)
        for i in range(0, X.shape[0], n_chunk):
            end = min(i + n_chunk, X.shape[0])

            _Ai = rbf_affinity(X[i:end].to(device), nystrom_X, gamma=gamma)
            _Ai, _indices = keep_topk_per_row(_Ai, n_neighbors)  # (n, n_neighbors)
            _Di = D[_indices].sum(1)
            _Ai = _Ai / _Di[:, None]

            weights = _Ai[..., None]  # (n, n_neighbors, 1)
            neighbors = nystrom_out[_indices.flatten()]
            neighbors = neighbors.reshape(-1, n_neighbors, nystrom_out.shape[-1])  # (n, n_neighbors, d)
            out = weights * neighbors  # (n, n_neighbors, d)
            out = out.sum(dim=1)  # (n, d)

            if config.move_output_to_cpu and not track_grad:
                out = out.to("cpu")
            all_outs.append(out)

        all_outs = torch.cat(all_outs, dim=0)

        if return_indices:
            return all_outs, indices
        
        return all_outs






URLS = {
    "dinov3_vits16": "https://huggingface.co/huzey/mydv3/resolve/master/dinov3_vits16_pretrain_lvd1689m-08c60483.pth",
    "dinov3_vits16plus": "https://huggingface.co/huzey/mydv3/resolve/master/dinov3_vits16plus_pretrain_lvd1689m-4057cbaa.pth",
    "dinov3_vitb16": "https://huggingface.co/huzey/mydv3/resolve/master/dinov3_vitb16_pretrain_lvd1689m-73cec8be.pth",
    "dinov3_vitl16": "https://huggingface.co/huzey/mydv3/resolve/master/dinov3_vitl16_pretrain_lvd1689m-8aa4cbdd.pth",
    "dinov3_vith16plus": "https://huggingface.co/huzey/mydv3/resolve/master/dinov3_vith16plus_pretrain_lvd1689m-7c1da9a5.pth",
    "dinov3_vitl16_sat493m": "https://huggingface.co/huzey/mydv3/resolve/master/dinov3_vitl16_pretrain_sat493m-eadcf0ff.pth",
    "dinov3_vitl16_dinotxt": "https://huggingface.co/huzey/mydv3/resolve/master/dinov3_vitl16_dinotxt_vision_head_and_text_encoder-a442d8f5.pth",
}

import torch
from torch import nn


class Dinov3Backbone(nn.Module):
    def __init__(self, config="dinov3_vitl16"):
        super().__init__()
        if config == "dinov3_vitl16_sat493m":
            config = "dinov3_vitl16"
        dinov3 = torch.hub.load("facebookresearch/dinov3", config, weights=URLS[config])
        self.model = dinov3
        
        self.keep_idx = None

    def forward(self, x: torch.Tensor):
        out = self.model.get_intermediate_layers(x, reshape=True)[0]  # b, c, h, w
        if self.keep_idx is None:
            self.keep_idx = self.remove_high_variance_channels(out)
        out = out[:, self.keep_idx, :, :]
        return out
    
    def remove_high_variance_channels(self, out: torch.Tensor, n_remove: int = 8):
        _out = out.permute(0, 2, 3, 1)
        _out = _out.reshape(-1, _out.shape[-1])
        var = torch.var(_out, dim=0)
        var_sorted_idx = torch.argsort(var, descending=True)
        keep_idx = var_sorted_idx[n_remove:]
        return keep_idx



# Mod from torch/_lowrank.py
# because QR and SVD is not implemented for MPS (apple silicon)

# TODO: remove this file after torch have updated MPS support

"""Implement various linear algebra algorithms for low rank matrices."""
from typing import Optional

import torch
from torch import _linalg_utils as _utils, Tensor
from torch.overrides import handle_torch_function, has_torch_function


def qr_fallback_cpu(X: torch.Tensor):
    device = X.device
    if device.type == 'mps':
        X = X.cpu()
        Q = torch.linalg.qr(X).Q
        Q = Q.to(device)
        return Q
    return torch.linalg.qr(X).Q


def svd_fallback_cpu(X: torch.Tensor):
    device = X.device
    if device.type == 'mps':
        X = X.cpu()
        U, S, V = torch.linalg.svd(X, full_matrices=False)
        U = U.to(device)
        S = S.to(device)
        V = V.to(device)
        return U, S, V
    return torch.linalg.svd(X, full_matrices=False)


def get_approximate_basis(
        A: Tensor,
        q: int,
        niter: Optional[int] = 2,
        M: Optional[Tensor] = None,
) -> Tensor:
    """Return tensor :math:`Q` with :math:`q` orthonormal columns such
    that :math:`Q Q^H A` approximates :math:`A`. If :math:`M` is
    specified, then :math:`Q` is such that :math:`Q Q^H (A - M)`
    approximates :math:`A - M`. without instantiating any tensors
    of the size of :math:`A` or :math:`M`.

    .. note:: The implementation is based on the Algorithm 4.4 from
              Halko et al., 2009.

    .. note:: For an adequate approximation of a k-rank matrix
              :math:`A`, where k is not known in advance but could be
              estimated, the number of :math:`Q` columns, q, can be
              chosen according to the following criteria: in general,
              :math:`k <= q <= min(2*k, m, n)`. For large low-rank
              matrices, take :math:`q = k + 5..10`.  If k is
              relatively small compared to :math:`min(m, n)`, choosing
              :math:`q = k + 0..2` may be sufficient.

    .. note:: To obtain repeatable results, reset the seed for the
              pseudorandom number generator

    Args::
        A (Tensor): the input tensor of size :math:`(*, m, n)`

        q (int): the dimension of subspace spanned by :math:`Q`
                 columns.

        niter (int, optional): the number of subspace iterations to
                               conduct; ``niter`` must be a
                               nonnegative integer. In most cases, the
                               default value 2 is more than enough.

        M (Tensor, optional): the input tensor's mean of size
                              :math:`(*, m, n)`.

    References::
        - Nathan Halko, Per-Gunnar Martinsson, and Joel Tropp, Finding
          structure with randomness: probabilistic algorithms for
          constructing approximate matrix decompositions,
          arXiv:0909.4061 [math.NA; math.PR], 2009 (available at
          `arXiv <http://arxiv.org/abs/0909.4061>`_).
    """

    niter = 2 if niter is None else niter
    dtype = _utils.get_floating_dtype(A) if not A.is_complex() else A.dtype
    matmul = _utils.matmul

    R = torch.randn(A.shape[-1], q, dtype=dtype, device=A.device)

    # The following code could be made faster using torch.geqrf + torch.ormqr
    # but geqrf is not differentiable

    X = matmul(A, R)
    if M is not None:
        X = X - matmul(M, R)
    Q = qr_fallback_cpu(X)
    for _ in range(niter):
        X = matmul(A.mH, Q)
        if M is not None:
            X = X - matmul(M.mH, Q)
        Q = qr_fallback_cpu(X)
        X = matmul(A, Q)
        if M is not None:
            X = X - matmul(M, Q)
        Q = qr_fallback_cpu(X)
    return Q


def svd_lowrank(
        A: Tensor,
        q: Optional[int] = 6,
        niter: Optional[int] = 2,
        M: Optional[Tensor] = None,
) -> tuple[Tensor, Tensor, Tensor]:
    r"""Return the singular value decomposition ``(U, S, V)`` of a matrix,
    batches of matrices, or a sparse matrix :math:`A` such that
    :math:`A \approx U \operatorname{diag}(S) V^{\text{H}}`. In case :math:`M` is given, then
    SVD is computed for the matrix :math:`A - M`.

    .. note:: The implementation is based on the Algorithm 5.1 from
              Halko et al., 2009.

    .. note:: For an adequate approximation of a k-rank matrix
              :math:`A`, where k is not known in advance but could be
              estimated, the number of :math:`Q` columns, q, can be
              chosen according to the following criteria: in general,
              :math:`k <= q <= min(2*k, m, n)`. For large low-rank
              matrices, take :math:`q = k + 5..10`.  If k is
              relatively small compared to :math:`min(m, n)`, choosing
              :math:`q = k + 0..2` may be sufficient.

    .. note:: This is a randomized method. To obtain repeatable results,
              set the seed for the pseudorandom number generator

    .. note:: In general, use the full-rank SVD implementation
              :func:`torch.linalg.svd` for dense matrices due to its 10x
              higher performance characteristics. The low-rank SVD
              will be useful for huge sparse matrices that
              :func:`torch.linalg.svd` cannot handle.

    Args::
        A (Tensor): the input tensor of size :math:`(*, m, n)`

        q (int, optional): a slightly overestimated rank of A.

        niter (int, optional): the number of subspace iterations to
                               conduct; niter must be a nonnegative
                               integer, and defaults to 2

        M (Tensor, optional): the input tensor's mean of size
                              :math:`(*, m, n)`, which will be broadcasted
                              to the size of A in this function.

    References::
        - Nathan Halko, Per-Gunnar Martinsson, and Joel Tropp, Finding
          structure with randomness: probabilistic algorithms for
          constructing approximate matrix decompositions,
          arXiv:0909.4061 [math.NA; math.PR], 2009 (available at
          `arXiv <https://arxiv.org/abs/0909.4061>`_).

    """
    if not torch.jit.is_scripting():
        tensor_ops = (A, M)
        if not set(map(type, tensor_ops)).issubset(
                (torch.Tensor, type(None))
        ) and has_torch_function(tensor_ops):
            return handle_torch_function(
                svd_lowrank, tensor_ops, A, q=q, niter=niter, M=M
            )
    return _svd_lowrank(A, q=q, niter=niter, M=M)


def _svd_lowrank(
        A: Tensor,
        q: Optional[int] = 6,
        niter: Optional[int] = 2,
        M: Optional[Tensor] = None,
) -> tuple[Tensor, Tensor, Tensor]:
    # Algorithm 5.1 in Halko et al., 2009

    q = 6 if q is None else q
    m, n = A.shape[-2:]
    matmul = _utils.matmul
    if M is not None:
        M = M.broadcast_to(A.size())

    # Assume that A is tall
    if m < n:
        A = A.mH
        if M is not None:
            M = M.mH

    Q = get_approximate_basis(A, q, niter=niter, M=M)
    B = matmul(Q.mH, A)
    if M is not None:
        B = B - matmul(Q.mH, M)
    U, S, Vh = svd_fallback_cpu(B)
    V = Vh.mH
    U = Q.matmul(U)

    if m < n:
        U, V = V, U

    return U, S, V



"""Utilities for managing PyTorch gradient computation state."""

import torch
from contextlib import contextmanager


@contextmanager
def set_grad_enabled(enabled: bool):
    """Context manager to temporarily set gradient computation mode.
    
    This context manager allows you to control gradient computation for a block
    of code, and automatically restores the previous gradient state when exiting
    the context.
    
    Args:
        enabled (bool): If True, enables gradient tracking within the context.
                        If False, disables gradient tracking within the context.
    
    Yields:
        None
        
    Examples:
        >>> import torch
        >>> from ncut_pytorch.utils.grad import set_grad_enabled
        >>> 
        >>> # Disable gradients for inference
        >>> with set_grad_enabled(False):
        ...     result = model(input_tensor)
        >>> 
        >>> # Enable gradients for training
        >>> with set_grad_enabled(True):
        ...     loss = criterion(model(input_tensor), target)
        ...     loss.backward()
    """
    prev_grad_state = torch.is_grad_enabled()
    torch.set_grad_enabled(enabled)
    try:
        yield
    finally:
        torch.set_grad_enabled(prev_grad_state)

